
import struct
from threading import Thread
from socket import socket, AF_INET, SOCK_DGRAM


class DEFAULT:
  port = 4213
  header = b'fxCyanF'

  listenAddr = '0.0.0.0'


class Sender:

  _socket = socket(AF_INET, SOCK_DGRAM)

  def __init__(self, host: str, port: int = DEFAULT.port):
    self._host = host
    self._port = port

  @staticmethod
  def _convertData(data):
    channels = []

    for item in data:
      f = float(item)

      if f > 1.0: f = 1.0
      if f < 0.0: f = 0.0

      channels += [f]

    return struct.pack('f' * len(channels), *channels)

  @classmethod
  def _sendUDP(cls, data: bytearray, host: str, port: int = DEFAULT.port):
    try:
      cls._socket.sendto(data, (host, port))
    except:
      # TODO: logging…?
      pass

  @classmethod
  def sendTo(cls, channels: list, host: str, port: int = DEFAULT.port):
    data = DEFAULT.header
    data += Sender._convertData(channels)

    cls._sendUDP(data, host, port)

  def send(self, channels: list):
    Sender.sendTo(channels, self._host, self._port)


class Receiver:

  def __init__(self, host: str = DEFAULT.listenAddr, port: int = DEFAULT.port):

    self._socket = socket(AF_INET, SOCK_DGRAM)
    self._socket.bind((host, port))

    self._callbacks = Set()

    t = Thread(target=self._recvLoop)
    t.daemon = True
    t.start()

  def _recvLoop(self):

    header = DEFAULT.header
    headerLen = len(header)

    while True:
      data = self._socket.recv(2048)

      if data.startswith(header):
        data = data[ headerLen : ]
      else:
        continue
      
      if len(data) % 4 != 0:
        # TODO: logging…?
        continue

      _chs = struct.unpack('f' * len(data), data)

      channels = []
      for ch in _chs:
        if ch > 1.0: ch = 1.0
        if ch < 0.0: ch = 0.0
        channels += [ch]

      for callback in self._callbacks:
        try:
          callback(channels)
        except:
          # TODO: logging…?
          pass

